package broker_test

import (
	"errors"
	"fmt"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/cloud_foundry_client"
)

var _ = Describe("Upgrade", func() {
	var (
		upgradeOperationData     broker.OperationData
		instanceID               string
		serviceDeploymentName    string
		logger                   *log.Logger
		expectedPreviousManifest []byte
		boshTaskID               int
		redeployErr              error
	)

	BeforeEach(func() {
		instanceID = "some-instance"
		serviceDeploymentName = deploymentName(instanceID)
		expectedPreviousManifest = []byte("old-manifest-fetched-from-bosh")
		boshTaskID = 876
	})

	JustBeforeEach(func() {
		logger = loggerFactory.NewWithRequestID()
		upgradeOperationData, redeployErr = b.Upgrade(instanceID, logger)
	})

	Context("when the deployment goes well", func() {
		BeforeEach(func() {
			boshClient.GetDeploymentReturns(expectedPreviousManifest, true, nil)
			fakeDeployer.UpgradeReturns(boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("does not error", func() {
			Expect(redeployErr).NotTo(HaveOccurred())
		})

		It("deploys with the new planID", func() {
			Expect(fakeDeployer.CreateCallCount()).To(Equal(0))
			Expect(fakeDeployer.UpgradeCallCount()).To(Equal(1))
			Expect(fakeDeployer.UpdateCallCount()).To(Equal(0))
			actualDeploymentName,
				actualPlanID,
				actualPreviousPlanID,
				actualBoshContextID,
				_ := fakeDeployer.UpgradeArgsForCall(0)
			Expect(actualPlanID).To(Equal(existingPlanID))
			Expect(actualDeploymentName).To(Equal(broker.InstancePrefix + instanceID))
			oldPlanIDCopy := existingPlanID
			Expect(actualPreviousPlanID).To(Equal(&oldPlanIDCopy))
			Expect(actualBoshContextID).To(BeEmpty())
		})
	})

	Context("when there is a previous deployment for the service instance", func() {
		BeforeEach(func() {
			boshClient.GetDeploymentReturns([]byte("crusty-old-manifest"), true, nil)
			fakeDeployer.UpgradeReturns(boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns the task ID for the upgrade task", func() {
			Expect(upgradeOperationData.BoshTaskID).To(Equal(boshTaskID))
		})

		It("returns an empty context id", func() {
			Expect(upgradeOperationData.BoshContextID).To(BeEmpty())
		})

		It("returns the plan id", func() {
			Expect(upgradeOperationData.PlanID).To(Equal(existingPlanID))
		})

		It("returns the operation type", func() {
			Expect(upgradeOperationData.OperationType).To(Equal(broker.OperationTypeUpgrade))
		})

		It("fetches correct instance", func() {
			Expect(cfClient.GetInstanceStateCallCount()).To(Equal(1))
			actualInstanceID, _ := cfClient.GetInstanceStateArgsForCall(0)
			Expect(actualInstanceID).To(Equal(instanceID))
		})

		It("logs with a request ID", func() {
			Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} upgrading instance`))
		})

		Context("and there is an operation in progress on the instance", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: existingPlanID, OperationInProgress: true}, nil)
			})

			It("returns an OperationInProgressError", func() {
				Expect(redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
			})
		})

		Context("and post-deploy errand is configured", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: postDeployErrandPlanID}, nil)
			})

			It("deploys with a context id", func() {
				_, _, _, contextID, _ := fakeDeployer.UpgradeArgsForCall(0)
				Expect(contextID).ToNot(BeEmpty())
			})

			It("returns the correct context id", func() {
				_, _, _, contextID, _ := fakeDeployer.UpgradeArgsForCall(0)
				Expect(upgradeOperationData.BoshContextID).To(Equal(contextID))
			})
		})
	})

	Context("when the service instance cannot be found", func() {
		BeforeEach(func() {
			boshClient.GetDeploymentReturns([]byte("crusty-old-manifest"), true, nil)
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{}, cloud_foundry_client.ResourceNotFoundError{})
		})

		It("returns an error", func() {
			Expect(redeployErr).To(HaveOccurred())
		})
	})

	Context("when the instance state cannot be retrieved from Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetDeploymentReturns([]byte("crusty-old-manifest"), true, nil)
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{}, errors.New("no instance for you"))
		})

		It("returns the error to the cf user", func() {
			Expect(redeployErr).To(MatchError(ContainSubstring("no instance for you")))
		})
	})

	Context("when the plan cannot be found", func() {
		var planID string

		BeforeEach(func() {
			planID = "non-existent-plan-id"
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: planID}, nil)
			boshClient.GetDeploymentReturns([]byte("crusty-old-manifest"), true, nil)
		})
		It("returns a user displayable error", func() {
			Expect(redeployErr).To(MatchError(ContainSubstring(fmt.Sprintf("plan %s not found", planID))))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error: finding plan ID %s", planID)))
		})

		It("does not redeploy", func() {
			Expect(boshClient.GetDeploymentCallCount()).To(BeZero())
			Expect(fakeDeployer.UpgradeCallCount()).To(BeZero())
		})
	})

	Context("when there is a task in progress on the instance", func() {
		BeforeEach(func() {
			boshClient.GetDeploymentReturns([]byte("crusty-old-manifest"), true, nil)
			fakeDeployer.UpgradeReturns(0, []byte{}, broker.NewOperationInProgressError(fmt.Errorf("deployment %s is still in progress", deploymentName(instanceID))))
			cfClient.GetInstanceStateReturns(cloud_foundry_client.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns an OperationInProgressError", func() {
			Expect(redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
		})
	})
})
