package task

import (
	"errors"
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

//go:generate counterfeiter -o fakes/fake_bosh_client.go . BoshClient
type BoshClient interface {
	Deploy(manifest []byte, contextID string, logger *log.Logger) (int, error)
	GetTasks(deploymentName string, logger *log.Logger) (boshclient.BoshTasks, error)
	GetDeployment(name string, logger *log.Logger) ([]byte, bool, error)
}

//go:generate counterfeiter -o fakes/fake_manifest_generator.go . ManifestGenerator
type ManifestGenerator interface {
	GenerateManifest(deploymentName, planID string, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, logger *log.Logger) (BoshManifest, error)
}

//go:generate counterfeiter -o fakes/fake_feature_flags.go . FeatureFlags
type FeatureFlags interface {
	CFUserTriggeredUpgrades() bool
}

type deployer struct {
	boshClient        BoshClient
	manifestGenerator ManifestGenerator
	featureFlags      FeatureFlags
}

func NewDeployer(
	boshClient BoshClient,
	manifestGenerator ManifestGenerator,
	featureFlags FeatureFlags,
) deployer {
	return deployer{
		boshClient:        boshClient,
		manifestGenerator: manifestGenerator,
		featureFlags:      featureFlags,
	}
}

func (d deployer) Create(
	deploymentName,
	planID string,
	requestParams map[string]interface{},
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {

	err := d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	return d.doDeploy(
		deploymentName,
		planID,
		broker.OperationTypeCreate,
		requestParams,
		nil,
		nil,
		boshContextID,
		logger,
	)
}

func (d deployer) Upgrade(
	deploymentName,
	planID string,
	previousPlanID *string,
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {

	oldManifest, err := d.getDeploymentManifest(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	err = d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	return d.doDeploy(
		deploymentName,
		planID,
		broker.OperationTypeUpgrade,
		nil,
		oldManifest,
		previousPlanID,
		boshContextID,
		logger,
	)
}

func (d deployer) Update(
	deploymentName,
	planID string,
	requestParams map[string]interface{},
	previousPlanID *string,
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {

	oldManifest, err := d.getDeploymentManifest(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	err = d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	parameters := parametersFromRequest(requestParams)
	applyChanges, err := d.validatedApplyChanges(parameters)
	if err != nil {
		return 0, nil, err
	}

	if applyChanges {
		err = d.assertCanApplyChanges(parameters, planID, previousPlanID)
		if err != nil {
			return 0, nil, err
		}
	}

	pendingChanges, pendingChangesErr := d.hasPendingChanges(
		deploymentName,
		previousPlanID,
		oldManifest,
		logger,
	)
	if pendingChangesErr != nil {
		return 0, nil, pendingChangesErr
	}

	if pendingChanges && !applyChanges {
		err = errors.New("pending changes detected")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return 0, nil, broker.NewPendingChangesError(err)
		} else {
			return 0, nil, broker.NewApplyChangesDisabledError(err)
		}
	}

	return d.doDeploy(
		deploymentName,
		planID,
		broker.OperationTypeUpdate,
		requestParams,
		oldManifest,
		previousPlanID,
		boshContextID,
		logger,
	)
}

func (d deployer) getDeploymentManifest(deploymentName string, logger *log.Logger) ([]byte, error) {
	oldManifest, found, err := d.boshClient.GetDeployment(deploymentName, logger)

	if err != nil {
		return nil, err
	}

	if !found {
		return nil, broker.NewDeploymentNotFoundError(fmt.Errorf("bosh deployment '%s' not found", deploymentName))
	}

	return oldManifest, nil
}

func (d deployer) assertNoOperationsInProgress(deploymentName string, logger *log.Logger) error {
	tasks, err := d.boshClient.GetTasks(deploymentName, logger)
	if err != nil {
		return fmt.Errorf("error getting tasks for deployment %s: %s\n", deploymentName, err)
	}

	if incompleteTasks := tasks.IncompleteTasks(); len(incompleteTasks) != 0 {
		userError := errors.New("An operation is in progress for your service instance. Please try again later.")
		operatorError := broker.NewOperationInProgressError(
			fmt.Errorf("deployment %s is still in progress: tasks %s\n",
				deploymentName,
				incompleteTasks.ToLog()),
		)
		return broker.NewDisplayableError(userError, operatorError)
	}

	return nil
}

func parametersFromRequest(requestParams map[string]interface{}) map[string]interface{} {
	parameters, ok := requestParams["parameters"].(map[string]interface{})
	if !ok {
		return nil
	}

	return parameters
}

func (d deployer) validatedApplyChanges(parameters map[string]interface{}) (bool, error) {
	if parameters["apply-changes"] == nil {
		return false, nil
	}

	applyChanges, ok := parameters["apply-changes"].(bool)
	if !ok {
		err := errors.New("update called with apply-changes set to non-boolean")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return false, broker.NewPendingChangesError(err)
		} else {
			return false, broker.NewApplyChangesNotPermittedError(err)
		}
	}

	delete(parameters, "apply-changes")

	return applyChanges, nil
}

func (d deployer) assertCanApplyChanges(
	parameters map[string]interface{},
	planID string,
	previousPlanID *string,
) error {
	if !d.featureFlags.CFUserTriggeredUpgrades() {
		return broker.NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))
	}

	if previousPlanID != nil && planID != *previousPlanID {
		err := errors.New("update called with apply-changes and a plan change")
		return broker.NewPendingChangesError(err)
	}

	if len(parameters) > 0 {
		err := errors.New("update called with apply-changes and arbitrary parameters set")
		return broker.NewPendingChangesError(err)
	}

	return nil
}

func (d deployer) hasPendingChanges(
	deploymentName string,
	previousPlanID *string,
	oldManifest BoshManifest,
	logger *log.Logger,
) (bool, error) {
	// generate a manifest without params and the previously used plan ID
	regeneratedManifest, err := d.manifestGenerator.GenerateManifest(deploymentName, *previousPlanID, map[string]interface{}{}, oldManifest, previousPlanID, logger)
	if err != nil {
		return false, err
	}

	// check whether the manifest generated by the adapter is the same we previously deployed
	manifestsEqual, err := regeneratedManifest.Equals(oldManifest)
	if err != nil {
		return false, fmt.Errorf("error detecting change in manifest: %s", err)
	}

	return !manifestsEqual, nil
}

func (d deployer) doDeploy(
	deploymentName,
	planID string,
	operationType broker.OperationType,
	requestParams map[string]interface{},
	oldManifest []byte,
	previousPlanID *string,
	boshContextID string,
	logger *log.Logger,
) (int, []byte, error) {
	manifest, err := d.manifestGenerator.GenerateManifest(
		deploymentName,
		planID,
		requestParams,
		oldManifest,
		previousPlanID,
		logger,
	)
	if err != nil {
		return 0, nil, err
	}

	boshTaskID, err := d.boshClient.Deploy(manifest, boshContextID, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	logger.Printf("Bosh task ID for %s deployment %s is %d\n", operationType, deploymentName, boshTaskID)

	return boshTaskID, manifest, nil
}
