package upgrader

import (
	"encoding/json"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

//go:generate counterfeiter -o fakes/fake_listener.go . Listener
type Listener interface {
	Starting()
	InstancesToUpgrade(instances []string)
	InstanceUpgradeStarting(instance string, index, totalInstances int)
	InstanceUpgradeStartResult(status UpgradeStartResult)
	InstanceUpgraded(instance string, result string)
	WaitingFor(instance string, boshTaskId int)
	Progress(pollingInterval time.Duration, orphanCount, upgradedCount, upgradesLeftCount, deletedCount int)
	Finished(orphanCount, upgradedCount, deletedCount int)
}

type upgrader struct {
	brokerUsername  string
	brokerPassword  string
	brokerUrl       string
	pollingInterval time.Duration
	listener        Listener
}

type UpgradeStartResult string

const (
	ResultSuccess  UpgradeStartResult = "success"
	ResultRetry    UpgradeStartResult = "retry"
	ResultNotFound UpgradeStartResult = "not found"
	ResultOrphan   UpgradeStartResult = "orphan"
	ResultError    UpgradeStartResult = "error"
)

func New(brokerUsername, brokerPassword, brokerUrl string, pollingInterval int, listener Listener) upgrader {
	return upgrader{
		brokerUsername:  brokerUsername,
		brokerPassword:  brokerPassword,
		brokerUrl:       brokerUrl,
		pollingInterval: time.Duration(pollingInterval) * time.Second,
		listener:        listener,
	}
}

func (u upgrader) Upgrade() error {
	var (
		upgradedCount, orphansCount int
		deletedInstances            int
	)

	u.listener.Starting()

	instancesToUpgrade, err := u.getServiceInstances()
	numberOfInstances := len(instancesToUpgrade)

	if err != nil {
		return fmt.Errorf("error getting service instances: %s\n", err)
	}

	u.listener.InstancesToUpgrade(instancesToUpgrade)

	for len(instancesToUpgrade) != 0 {
		upgradedInstancesCount, orphanInstancesCount, retryInstances, err := u.upgradeInstances(instancesToUpgrade)
		if err != nil {
			return err
		}

		upgradedCount += upgradedInstancesCount
		orphansCount += orphanInstancesCount
		instancesToUpgrade = retryInstances
		deletedInstances = numberOfInstances - (upgradedCount + orphansCount + len(retryInstances))

		u.listener.Progress(u.pollingInterval, orphansCount, upgradedCount, len(instancesToUpgrade), deletedInstances)
		if len(instancesToUpgrade) > 0 {
			time.Sleep(u.pollingInterval)
		}
	}

	u.listener.Finished(orphansCount, upgradedCount, deletedInstances)

	return nil
}

func (u upgrader) upgradeInstances(instances []string) (int, int, []string, error) {
	var (
		upgradedCount, orphanCount int
		idsToRetry                 []string
	)

	instanceCount := len(instances)
	for i, instance := range instances {
		u.listener.InstanceUpgradeStarting(instance, i, instanceCount)
		operationData, result, err := u.upgradeInstance(instance)
		if err != nil {
			return 0, 0, nil, fmt.Errorf(
				"Upgrade failed for service instance %s: %s\n", instance, err,
			)
		}

		switch result {
		case ResultOrphan:
			orphanCount++
		case ResultRetry:
			idsToRetry = append(idsToRetry, instance)
		case ResultSuccess:
			if err := u.pollLastOperation(instance, operationData); err != nil {
				u.listener.InstanceUpgraded(instance, "failure") // TODO This error should just propagate up, without logging
				return 0, 0, nil, err
			}
			u.listener.InstanceUpgraded(instance, string(ResultSuccess))
			upgradedCount++
		}
	}

	return upgradedCount, orphanCount, idsToRetry, nil
}

func (u upgrader) pollLastOperation(instance string, opData broker.OperationData) error {
	u.listener.WaitingFor(instance, opData.BoshTaskID)

	for {
		time.Sleep(u.pollingInterval)

		lastOperation, err := u.getLastOperation(instance, opData)
		if err != nil {
			return fmt.Errorf("error getting last operation: %s\n", err)
		}

		switch lastOperation.State {
		case brokerapi.Failed:
			return fmt.Errorf("Upgrade failed for service instance %s: bosh task id %d: %s", instance, opData.BoshTaskID, lastOperation.Description)
		case brokerapi.Succeeded:
			return nil
		}
	}
}

func (u upgrader) getServiceInstances() ([]string, error) {
	var instances []mgmtapi.Instance
	if err := u.makeRequest("/mgmt/service_instances", &instances); err != nil {
		return nil, err
	}

	var instanceIDs []string
	for _, instance := range instances {
		instanceIDs = append(instanceIDs, instance.InstanceID)
	}
	return instanceIDs, nil
}

func (u upgrader) upgradeInstance(instance string) (broker.OperationData, UpgradeStartResult, error) {
	request, err := http.NewRequest("PATCH", fmt.Sprintf("%s%s", u.brokerUrl, "/mgmt/service_instances/"+instance), nil)
	if err != nil {
		return broker.OperationData{}, ResultError, err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return broker.OperationData{}, ResultError, err
	}
	defer response.Body.Close()

	logAndReturn := func(od broker.OperationData, r UpgradeStartResult) (broker.OperationData, UpgradeStartResult, error) {
		u.listener.InstanceUpgradeStartResult(r)
		return od, r, nil
	}

	switch response.StatusCode {
	case http.StatusAccepted:
		var operationData broker.OperationData
		if err = json.NewDecoder(response.Body).Decode(&operationData); err != nil {
			return broker.OperationData{}, ResultError, fmt.Errorf("can't parse response %s", err)
		}
		return logAndReturn(operationData, ResultSuccess)
	case http.StatusNotFound:
		return logAndReturn(broker.OperationData{}, ResultNotFound)
	case http.StatusGone:
		return logAndReturn(broker.OperationData{}, ResultOrphan)
	case http.StatusConflict:
		return logAndReturn(broker.OperationData{}, ResultRetry)
	}

	var errorResponse brokerapi.ErrorResponse
	if err := json.NewDecoder(response.Body).Decode(&errorResponse); err != nil {
		return broker.OperationData{}, ResultError, fmt.Errorf("can't parse response %s", err)
	}

	return broker.OperationData{}, ResultError, fmt.Errorf(
		"unexpected status code: %d. description: %s", response.StatusCode, errorResponse.Description,
	)
}

func (u upgrader) getLastOperation(instance string, opData broker.OperationData) (brokerapi.LastOperation, error) {
	opDataJSON, err := json.Marshal(opData)
	if err != nil {
		return brokerapi.LastOperation{}, err
	}

	operationQueryStr := url.QueryEscape(string(opDataJSON))

	var lastOperation brokerapi.LastOperation
	if err := u.makeRequest("/v2/service_instances/"+instance+"/last_operation?operation="+operationQueryStr, &lastOperation); err != nil {
		return brokerapi.LastOperation{}, err
	}

	return lastOperation, nil
}

func (u upgrader) makeRequest(requestPath string, responseData interface{}) error {
	request, err := http.NewRequest("GET", fmt.Sprintf("%s%s", u.brokerUrl, requestPath), nil)
	if err != nil {
		return err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return err
	}

	if response.StatusCode != 200 {
		return fmt.Errorf("response returned bad status: %s", response.Status)
	}

	err = json.NewDecoder(response.Body).Decode(responseData)
	if err != nil {
		return err
	}

	return nil
}
