package upgrader_test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"time"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/onsi/gomega/ghttp"
	"github.com/onsi/gomega/types"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
	"github.com/pivotal-cf/on-demand-service-broker/upgrader"
	"github.com/pivotal-cf/on-demand-service-broker/upgrader/fakes"
)

var _ = Describe("Upgrader", func() {
	const (
		zeroSeconds       = time.Duration(0) * time.Second
		brokerUsername    = "username"
		brokerPassword    = "password"
		pollingInterval   = 0
		serviceInstanceId = "serviceInstanceId"
		upgradeTaskID     = 412
	)
	lastOperationSucceeded := brokerapi.LastOperation{State: brokerapi.Succeeded}
	lastOperationInProgress := brokerapi.LastOperation{State: brokerapi.InProgress}

	var (
		server          *ghttp.Server
		brokerUrl       string
		actualErr       error
		logBuffer       *bytes.Buffer
		loggerFactory   *loggerfactory.LoggerFactory
		logger          *log.Logger
		listener        upgrader.Listener
		fakeListener    *fakes.FakeListener
		invalidResponse string
		output          string
	)

	BeforeEach(func() {
		server = ghttp.NewServer()
		brokerUrl = server.URL()
		logBuffer = new(bytes.Buffer)
		loggerFactory = loggerfactory.New(io.MultiWriter(GinkgoWriter, logBuffer), "upgrader-tool-tests", log.LstdFlags)
		logger = loggerFactory.NewWithRequestID()
		listener = upgrader.NewLoggingListener(logger)
		fakeListener = new(fakes.FakeListener)
	})

	AfterEach(func() {
		server.Close()
	})

	JustBeforeEach(func() {
		upgrader := upgrader.New(brokerUsername, brokerPassword, brokerUrl, pollingInterval, fakeListener)
		actualErr = upgrader.Upgrade()

		output = string(logBuffer.Bytes())
	})

	Describe("authentication", func() {
		Context("success", func() {
			BeforeEach(func() {
				server.AppendHandlers(
					ghttp.CombineHandlers(
						ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
						ghttp.RespondWithJSONEncoded(http.StatusOK, []mgmtapi.Instance{}),
						ghttp.VerifyBasicAuth(brokerUsername, brokerPassword),
					))
			})

			It("autenticates with the given credentials", func() {
				Expect(actualErr).NotTo(HaveOccurred())
			})
		})

		Context("failure", func() {
			BeforeEach(func() {
				server.AppendHandlers(ghttp.CombineHandlers(
					ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
					ghttp.RespondWithJSONEncoded(http.StatusUnauthorized, []mgmtapi.Instance{}),
				))
			})

			It("returns an error", func() {
				Expect(actualErr).To(HaveOccurred())
			})
		})
	})

	Context("upgrades one instance", func() {
		Context("and succeeds", func() {
			BeforeEach(func() {
				server.AppendHandlers(
					instancesRequestResponds(serviceInstanceId),
					upgradeRequestReponds(serviceInstanceId, http.StatusAccepted, upgradeResponse(upgradeTaskID)),
					lastOperationRequestReponds(serviceInstanceId, upgradeResponse(upgradeTaskID), lastOperationSucceeded),
				)
			})

			It("returns the list of successful upgrades", func() {
				hasReportedInstanceUpgradeStarted(fakeListener, serviceInstanceId, 0, 1)
				hasReportedInstanceUpgradeStartResult(fakeListener, upgrader.ResultSuccess)
				hasReportedUpgraded(fakeListener, serviceInstanceId)
				Expect(actualErr).NotTo(HaveOccurred())
			})
		})

		Describe("and fails, to get service instances", func() {
			var errMsg string
			JustBeforeEach(func() {
				if actualErr != nil {
					errMsg = actualErr.Error()
				}
			})

			Context("due to a bad status code", func() {
				BeforeEach(func() {
					server.AppendHandlers(ghttp.CombineHandlers(
						ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
						ghttp.RespondWith(http.StatusNotFound, "{}"),
					))
				})

				It("return an error", func() {
					Expect(errMsg).To(
						ContainSubstring("error getting service instances: response returned bad status: 404 Not Found"),
					)
				})
			})

			Context("due to a invalid url", func() {
				BeforeEach(func() {
					brokerUrl = "not a valid url"
				})

				It("returns an error", func() {
					Expect(errMsg).To(ContainSubstring(
						"error getting service instances: Get not%20a%20valid%20url/mgmt/service_instances: unsupported protocol scheme"),
					)
				})
			})

			Context("due to a timed out request", func() {
				BeforeEach(func() {
					longRequestHandler := func(w http.ResponseWriter, req *http.Request) {
						time.Sleep(1 * time.Second)
					}
					server.AppendHandlers(ghttp.CombineHandlers(
						ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
						longRequestHandler,
					))
				})

				It("returns an error", func() {
					Expect(errMsg).To(ContainSubstring("error getting service instances: EOF"))
				})
			})

			Context("due to a malformed response", func() {
				invalidResponse = "invalid response"
				BeforeEach(func() {
					server.AppendHandlers(ghttp.CombineHandlers(
						ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
						ghttp.RespondWith(http.StatusOK, invalidResponse),
					))
				})

				It("returns an error", func() {
					Expect(errMsg).To(ContainSubstring(
						"error getting service instances: invalid character 'i' looking for beginning of value"),
					)
				})
			})
		})
	})

	Context("upgrading an instance is not instant", func() {
		BeforeEach(func() {
			upgradeResponse := broker.OperationData{
				BoshTaskID:    upgradeTaskID,
				BoshContextID: "some-context-id",
				PlanID:        "some-plan-id",
				OperationType: broker.OperationTypeUpgrade,
			}

			server.AppendHandlers(
				instancesRequestResponds(serviceInstanceId),
				upgradeRequestReponds(serviceInstanceId, http.StatusAccepted, upgradeResponse),
				lastOperationRequestReponds(serviceInstanceId, upgradeResponse, lastOperationInProgress),
				lastOperationRequestReponds(serviceInstanceId, upgradeResponse, lastOperationInProgress),
				lastOperationRequestReponds(serviceInstanceId, upgradeResponse, lastOperationSucceeded),
			)
		})

		It("polls last operation until successful", func() {
			Expect(requestsMatching(server, ContainSubstring("/last_operation"))).To(Equal(3))
			Expect(actualErr).NotTo(HaveOccurred())
			hasReportedUpgraded(fakeListener, serviceInstanceId)
		})
	})

	Context("if the CF service instance has been deleted", func() {
		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstanceId),
				upgradeRequestReponds(serviceInstanceId, http.StatusNotFound, nil),
			)
		})

		It("ignores one instance", func() {
			Expect(actualErr).NotTo(HaveOccurred())

			hasReportedInstanceUpgradeStartResult(fakeListener, upgrader.ResultNotFound)
			hasReportedProgress(fakeListener, zeroSeconds, 0, 0, 0, 1)
			hasReportedFinished(fakeListener, 0, 0, 1)
		})
	})

	Context("if the bosh deployment cannot be found", func() {
		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstanceId),
				upgradeRequestReponds(serviceInstanceId, http.StatusGone, nil),
			)
		})

		It("detects one orphan instance", func() {
			Expect(actualErr).NotTo(HaveOccurred())

			hasReportedInstanceUpgradeStartResult(fakeListener, upgrader.ResultOrphan)
			hasReportedProgress(fakeListener, zeroSeconds, 1, 0, 0, 0)
			hasReportedFinished(fakeListener, 1, 0, 0)
		})
	})

	Context("A service is being held by another operation", func() {
		const serviceInstanceId = "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstanceId),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusAccepted, upgradeResponse(upgradeTaskID)),
				lastOperationRequestReponds(serviceInstanceId, upgradeResponse(upgradeTaskID), lastOperationSucceeded),
			)
		})

		It("retries upgrade until accepted", func() {
			Expect(actualErr).NotTo(HaveOccurred())

			Expect(requestsMatching(server, HaveSuffix("/mgmt/service_instances/"+serviceInstanceId))).To(Equal(4), "number of service requests")
			hasReportedInstanceUpgradeStartResult(
				fakeListener,
				upgrader.ResultRetry,
				upgrader.ResultRetry,
				upgrader.ResultRetry,
				upgrader.ResultSuccess,
			)
			hasReportedRetries(fakeListener, 1, 1, 1, 0)
			hasReportedFinished(fakeListener, 0, 1, 0)
		})
	})

	Context("A service is being deleted by another operation", func() {
		const serviceInstanceId = "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstanceId),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstanceId, http.StatusGone, nil),
			)
		})

		It("retries until orphan service instance detected", func() {
			Expect(actualErr).NotTo(HaveOccurred())
			Expect(requestsMatching(server, HaveSuffix("/mgmt/service_instances/"+serviceInstanceId))).To(Equal(4), "number of service requests")

			hasReportedRetries(fakeListener, 1, 1, 1, 0)
			hasReportedOrphans(fakeListener, 0, 0, 0, 1)
			hasReportedFinished(fakeListener, 1, 0, 0)
		})
	})

	Context("upgrades multiple instances", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		upgradeTaskID1 := 123
		upgradeTaskID2 := 456
		upgradeTaskID3 := 789

		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstance1, serviceInstance2, serviceInstance3),
				upgradeRequestReponds(serviceInstance1, http.StatusAccepted, upgradeResponse(upgradeTaskID1)),
				lastOperationRequestReponds(serviceInstance1, upgradeResponse(upgradeTaskID1), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusAccepted, upgradeResponse(upgradeTaskID2)),
				lastOperationRequestReponds(serviceInstance2, upgradeResponse(upgradeTaskID2), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance3, http.StatusAccepted, upgradeResponse(upgradeTaskID3)),
				lastOperationRequestReponds(serviceInstance3, upgradeResponse(upgradeTaskID3), lastOperationSucceeded),
			)
		})

		It("succesfully", func() {
			Expect(actualErr).NotTo(HaveOccurred())

			hasReportedStarting(fakeListener)
			hasReportedInstancesToUpgrade(fakeListener, serviceInstance1, serviceInstance2, serviceInstance3)
			hasReportedWaitingFor(fakeListener, map[string]int{serviceInstance1: upgradeTaskID1, serviceInstance2: upgradeTaskID2, serviceInstance3: upgradeTaskID3})
			hasReportedUpgraded(fakeListener, serviceInstance1, serviceInstance2, serviceInstance3)
			hasReportedProgress(fakeListener, zeroSeconds, 0, 3, 0, 0)
			hasReportedFinished(fakeListener, 0, 3, 0)
		})
	})

	Context("upgrades multiple instances, upgrade fails", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		upgradeTaskID1 := 123

		expectedErrorMessage := "everything went wrong"

		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstance1, serviceInstance2, serviceInstance3),
				upgradeRequestReponds(serviceInstance1, http.StatusAccepted, upgradeResponse(upgradeTaskID1)),
				lastOperationRequestReponds(serviceInstance1, upgradeResponse(upgradeTaskID1), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusInternalServerError, brokerapi.ErrorResponse{Description: expectedErrorMessage}),
			)
		})

		It("returns the error with the description from the http response", func() {
			message := fmt.Sprintf(
				"Upgrade failed for service instance %s: unexpected status code: %d. description: %s",
				serviceInstance2,
				http.StatusInternalServerError,
				expectedErrorMessage,
			)
			Expect(actualErr).To(MatchError(ContainSubstring(message)))
		})
	})

	Context("an operation fails when upgrading multiple instances", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		expectedErrorMessage := "everything went wrong"
		upgradeTaskID1 := 432
		upgradeTaskID2 := 987

		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstance1, serviceInstance2, serviceInstance3),
				upgradeRequestReponds(serviceInstance1, http.StatusAccepted, upgradeResponse(upgradeTaskID1)),
				lastOperationRequestReponds(serviceInstance1, upgradeResponse(upgradeTaskID1), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusAccepted, upgradeResponse(upgradeTaskID2)),
				lastOperationRequestReponds(
					serviceInstance2,
					upgradeResponse(upgradeTaskID2),
					brokerapi.LastOperation{State: brokerapi.Failed, Description: expectedErrorMessage}),
			)
		})

		It("reports the error from the server response", func() {
			Expect(actualErr).To(MatchError(fmt.Sprintf("Upgrade failed for service instance %s: bosh task id %d: %s", serviceInstance2, upgradeTaskID2, expectedErrorMessage)))

			hasReportedWaitingFor(fakeListener, map[string]int{serviceInstance1: upgradeTaskID1, serviceInstance2: upgradeTaskID2})
			hasReportedFailureFor(fakeListener, serviceInstance2)
		})
	})

	Context("With one instance orphaned", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		upgradeTaskID1 := 321
		upgradeTaskID3 := 1234

		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstance1, serviceInstance2, serviceInstance3),
				upgradeRequestReponds(serviceInstance1, http.StatusAccepted, upgradeResponse(upgradeTaskID1)),
				lastOperationRequestReponds(serviceInstance1, upgradeResponse(upgradeTaskID1), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusGone, nil),
				upgradeRequestReponds(serviceInstance3, http.StatusAccepted, upgradeResponse(upgradeTaskID3)),
				lastOperationRequestReponds(serviceInstance3, upgradeResponse(upgradeTaskID3), lastOperationSucceeded),
			)
		})

		It("upgrades multiple instances", func() {
			Expect(actualErr).NotTo(HaveOccurred())
			hasReportedFinished(fakeListener, 1, 2, 0)
		})
	})

	Context("upgrading multiple instances, where one is locked by bosh", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		upgradeTaskID1 := 4321
		upgradeTaskID2 := 9876
		upgradeTaskID3 := 6543

		BeforeEach(func() {
			server.AppendHandlers(
				instancesRequestResponds(serviceInstance1, serviceInstance2, serviceInstance3),
				upgradeRequestReponds(serviceInstance1, http.StatusAccepted, upgradeResponse(upgradeTaskID1)),
				lastOperationRequestReponds(serviceInstance1, upgradeResponse(upgradeTaskID1), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstance3, http.StatusAccepted, upgradeResponse(upgradeTaskID3)),
				lastOperationRequestReponds(serviceInstance3, upgradeResponse(upgradeTaskID3), lastOperationSucceeded),
				upgradeRequestReponds(serviceInstance2, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstance2, http.StatusConflict, nil),
				upgradeRequestReponds(serviceInstance2, http.StatusAccepted, upgradeResponse(upgradeTaskID2)),
				lastOperationRequestReponds(serviceInstance2, upgradeResponse(upgradeTaskID2), lastOperationSucceeded),
			)
		})

		It("retries until all upgraded", func() {
			Expect(actualErr).NotTo(HaveOccurred())

			Expect(requestsMatching(server, HaveSuffix("/mgmt/service_instances/"+serviceInstance2))).To(Equal(4), "number of service requests")
			hasReportedRetries(fakeListener, 1, 1, 1, 0)
			hasReportedFinished(fakeListener, 0, 3, 0)
		})
	})
})

func upgradeResponse(taskId int) broker.OperationData {
	return broker.OperationData{BoshTaskID: taskId, OperationType: broker.OperationTypeUpgrade}
}

func hasReportedStarting(fakeListener *fakes.FakeListener) {
	Expect(fakeListener.StartingCallCount()).To(Equal(1))
}

func hasReportedInstancesToUpgrade(fakeListener *fakes.FakeListener, instanceIds ...string) {
	Expect(fakeListener.InstancesToUpgradeCallCount()).To(Equal(1))
	Expect(fakeListener.InstancesToUpgradeArgsForCall(0)).To(Equal(instanceIds))
}

func hasReportedWaitingFor(fakeListener *fakes.FakeListener, instances map[string]int) {
	calls := fakeListener.WaitingForCallCount()
	Expect(calls).To(Equal(len(instances)))
	for i := 0; i < calls; i++ {
		instanceId, taskId := fakeListener.WaitingForArgsForCall(i)
		Expect(instances[instanceId]).To(Equal(taskId), "Task Id for "+instanceId)
	}
}

func hasReportedInstanceUpgradeStarted(fakeListener *fakes.FakeListener, expectedInstance string, expectedIndex, expectedTotalInstances int) {
	Expect(fakeListener.InstanceUpgradeStartingCallCount()).To(
		Equal(1), "instance upgrade started call count",
	)

	actualInstance,
		actualIndex,
		actualTotalInstances := fakeListener.InstanceUpgradeStartingArgsForCall(0)
	Expect(actualInstance).To(Equal(expectedInstance))
	Expect(actualIndex).To(Equal(expectedIndex), "expected index for instance upgrade started")
	Expect(actualTotalInstances).To(Equal(expectedTotalInstances), "expected total num of instances for instance upgrade started")
}

func hasReportedInstanceUpgradeStartResult(fakeListener *fakes.FakeListener, expectedStatuses ...upgrader.UpgradeStartResult) {
	Expect(fakeListener.InstanceUpgradeStartResultCallCount()).To(
		Equal(len(expectedStatuses)), "instance upgrade start result call count",
	)

	for i, expectedStatus := range expectedStatuses {
		Expect(fakeListener.InstanceUpgradeStartResultArgsForCall(i)).To(
			Equal(expectedStatus),
		)
	}
}

func hasReportedUpgraded(fakeListener *fakes.FakeListener, expectedInstanceIds ...string) {
	hasReportedUpgradeStates(fakeListener, "success", expectedInstanceIds...)
}

func hasReportedFailureFor(fakeListener *fakes.FakeListener, expectedInstanceIds ...string) {
	hasReportedUpgradeStates(fakeListener, "failure", expectedInstanceIds...)
}

func hasReportedUpgradeStates(fakeListener *fakes.FakeListener, expectedStatus string, expectedInstanceIds ...string) {
	upgraded := make([]string, 0)
	for i := 0; i < fakeListener.InstanceUpgradedCallCount(); i++ {
		id, status := fakeListener.InstanceUpgradedArgsForCall(i)
		if status == expectedStatus {
			upgraded = append(upgraded, id)
		}
	}
	Expect(upgraded).To(Equal(expectedInstanceIds), "status="+expectedStatus)
}

func hasReportedRetries(fakeListener *fakes.FakeListener, expectedRetryCounts ...int) {
	for i, expectedRetryCount := range expectedRetryCounts {
		_, _, _, toRetryCount, _ := fakeListener.ProgressArgsForCall(i)
		Expect(toRetryCount).To(Equal(expectedRetryCount), "Retry count: "+string(i))
	}
}

func hasReportedOrphans(fakeListener *fakes.FakeListener, expectedOrphanCounts ...int) {
	for i, expectedOrphanCount := range expectedOrphanCounts {
		_, orphanCount, _, _, _ := fakeListener.ProgressArgsForCall(i)
		Expect(orphanCount).To(Equal(expectedOrphanCount), "Orphan count: "+string(i))
	}
}

func hasReportedProgress(fakeListener *fakes.FakeListener, expectedInterval time.Duration, expectedOrphans, expectedUpgraded, expectedToRetry, expectedDeleted int) {
	Expect(fakeListener.ProgressCallCount()).To(Equal(1))
	pollingInterval, orphanCount, upgradedCount, toRetryCount, deletedCount := fakeListener.ProgressArgsForCall(0)
	Expect(pollingInterval).To(Equal(expectedInterval), "polling interval")
	Expect(orphanCount).To(Equal(expectedOrphans), "orphans")
	Expect(upgradedCount).To(Equal(expectedUpgraded), "upgraded")
	Expect(toRetryCount).To(Equal(expectedToRetry), "to retry")
	Expect(deletedCount).To(Equal(expectedDeleted), "deleted")
}

func hasReportedFinished(fakeListener *fakes.FakeListener, expectedOrphans, expectedUpgraded, expectedDeleted int) {
	Expect(fakeListener.FinishedCallCount()).To(Equal(1))
	orphanCount, upgradedCount, deletedCount := fakeListener.FinishedArgsForCall(0)
	Expect(orphanCount).To(Equal(expectedOrphans), "orphans")
	Expect(upgradedCount).To(Equal(expectedUpgraded), "upgraded")
	Expect(deletedCount).To(Equal(expectedDeleted), "deleted")
}

func requestsMatching(server *ghttp.Server, matcher types.GomegaMatcher) int {
	calls := 0
	for _, req := range server.ReceivedRequests() {
		if match, _ := matcher.Match(req.URL.String()); match {
			calls++
		}
	}
	return calls
}

func instancesRequestResponds(instanceIds ...string) http.HandlerFunc {
	instances := make([]mgmtapi.Instance, len(instanceIds))
	for i, id := range instanceIds {
		instances[i] = mgmtapi.Instance{InstanceID: id}
	}

	return ghttp.CombineHandlers(
		ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
		ghttp.RespondWithJSONEncoded(http.StatusOK, instances),
	)
}

func lastOperationRequestReponds(serviceInstanceId string, upgradeResponse broker.OperationData, lastOperation brokerapi.LastOperation) http.HandlerFunc {
	return ghttp.CombineHandlers(
		ghttp.VerifyRequest(
			"GET",
			"/v2/service_instances/"+serviceInstanceId+"/last_operation",
			"operation="+operationDataQuery(upgradeResponse)),
		ghttp.RespondWithJSONEncoded(http.StatusOK, lastOperation))
}

func upgradeRequestReponds(instanceId string, responseStatus int, response interface{}) http.HandlerFunc {
	return ghttp.CombineHandlers(
		ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+instanceId),
		ghttp.RespondWithJSONEncoded(responseStatus, response),
	)
}

func operationDataQuery(operationData broker.OperationData) string {
	operationDataBytes, err := json.Marshal(operationData)
	Expect(err).NotTo(HaveOccurred())
	return url.QueryEscape(string(operationDataBytes))
}
